@@ -69,9 +69,9 @@ class Circular():
         
         model = background((100, 100), {'vp':vp_back, 'vs':vs_back, 'rho':rho_back})        
         if not smoothing:  # Not m0
-            model['vp'] = add_circle(model['vp'], vp_circle, r=4, cx=35, cz=40)
-            model['vs'] = add_circle(model['vs'], vs_circle, r=4, cx=50, cz=50)
-            model['rho'] = add_circle(model['rho'], rho_circle, r=4, cx=65, cz=65)
+            model['vp'] = add_circle(model['vp'], vp_circle, r=6, cx=32, cz=32)
+            model['vs'] = add_circle(model['vs'], vs_circle, r=6, cx=50, cz=50)
+            model['rho'] = add_circle(model['rho'], rho_circle, r=6, cx=68, cz=68)
             if vintage == 2:  # Monitor model
                 model['vp'][25:30, 30: 41] -= 0.2 * model['vp'][25:30, 30: 41]
                
@@ -236,7 +236,10 @@ class ModelGenerator(Circular, Laminar):
     
     def marmousi(self, vintage, smoothing):
         path = os.path.dirname(__file__) + '/data/'
-    
+        
+        if os.path.exists(path) is False:
+            os.mkdir(path)
+
         target_path = path + "elastic-marmousi-model.tar.gz"
 
         models_url = "https://s3.amazonaws.com/open.source.geoscience/open_data/elastic-marmousi/elastic-marmousi-model.tar.gz"
@@ -443,10 +446,13 @@ def add_circle (model, circle_prop, r, cx, cz):
     return model
 
 
-def model_smoother(model, smooting_value):
+def model_smoother(model, smoothing_value, constant_surface=0, constant_left=0):
+    model0 = {}
     for params in model:
-        model[params] = gaussian_filter(model[params], smooting_value)
-    return model
+        model0[params] = gaussian_filter(model[params], smoothing_value)
+        model0[params][:constant_surface, :] = model[params][:constant_surface, :]
+        model0[params][:, :constant_left] = model[params][:, :constant_left]
+    return model0
 
 def pcs_perturbation(rho=None, prop_back=None, prop_circle=None, nz=100, nx=100, r=8, monitor=False):
     """
@@ -535,7 +541,7 @@ def model_resizing(model0,  bx=None, ex=None, bz=None, ez=None, ssr=(1, 1)):
         model[param] = interpolator(xi, zi)
         model[param] = model[param].astype(np.float32, order='C')
 
-        if bx:
+        if bx is not None:
             model[param] = model[param][bz:ez, bx:ex]
     return model
 
